''' 
Weighted Petri net with place and transition labels.

Allows explicit place or transition ids for simpler comparison, especially
during testing.

For an depth understanding or introduction to Petri Nets, see:
    - [Reference A]
    - [Reference B]
    - [Reference C]
'''

from collections.abc import Iterable
from typing import Union,FrozenSet
import xml.etree.ElementTree as ET
from uuid import uuid4

ENCODING='unicode'

# Candidate to move to a utility package
def steq(self,other):
    if type(other) is type(self):
        return self.__dict__ == other.__dict__
    return False

def verbosecmp(obj1:object,obj2:object) -> str:
    """
    This function produces a verbose statement about the given equality of
    the two objects.

    Returns a string statement about this equality.    
    """
    if obj1 == obj2:
        return "Same"
    result = ""
    if obj1.__dict__.keys() != obj2.__dict__.keys():
        ks1 = set(obj1.__dict__.keys())
        ks2 = set(obj2.__dict__.keys())
        result += f"Attributes not in both objects: {ks1 ^ ks2}\n"
    if obj1.__dict__ != obj2.__dict__:
        for k in obj1.__dict__:
            if k in obj2.__dict__ and obj1.__dict__[k] != obj2.__dict__[k] :
                result += f"Attribute {k} differs:\n"\
                        + f"  {obj1.__dict__[k]} vs\n  {obj2.__dict__[k]}\n"
    return result

class Place:
    """
    This a hashable and identifable place for a petri net.
    A place has a name and an identifier.
    """

    def __init__(self,name:str,pid:str=None):
        self._name = name
        # create an identifier.
        if (pid == None):
            self._pid = str(uuid4())
        else:
            self._pid = pid

    @property 
    def name(self) -> str:
        return self._name

    @property
    def pid(self) -> str:
        return self._pid

    @property
    def nodeId(self) -> str:
        return self._pid

    def __eq__(self,other) -> bool:
        if type(other) == type(self):
            return self.name == other.name and self.pid == other.pid
        return False

    def __hash__(self) -> int:
        return hash((self._name,self._pid))

    def __str__(self) -> str:
        return f'({self.name})'  if self._pid is None \
               else f'({self.name}({self._pid}))'

    def __repr__(self) -> str:
        return f'Place("{self.name}",pid="{self.pid}")'

class Transition:
    """
    This is hashable and identifable transition for a Petri net.
    A transition has a name, an identifier, a possible weight and can be silent.
    """

    def __init__(self,name:str,tid:str=None,weight:float=1.0,silent:bool=False):
        self._name = name
        # create an identifier.
        if (tid == None):
            self._tid = str(uuid4())
        else:
            self._tid = tid
        # add extras
        self._weight = weight
        self._silent = silent

    @property 
    def name(self) -> str:
        return self._name

    @property
    def silent(self) -> bool:
        return self._silent

    @property
    def tid(self) -> str:
        return self._tid

    @property
    def nodeId(self) -> str:
        return self._tid

    @property
    def weight(self) -> float:
        return self._weight

    @weight.setter
    def weight(self,value:float) -> None:
        self._weight = value

    def __eq__(self,other) -> bool:
        if type(self) == type(other):
            return self.name == other.name and self.tid == other.tid and  \
                   self.weight == other.weight and self._silent == other._silent
        return False

    def __hash__(self) -> int:
        return hash((self._name,self._tid,self._weight,self._silent))

    def __str__(self) -> str:
        return f'[{self.name} {self.weight}]'  if self._tid is None \
               else f'[{self.name}({self._tid}) {self.weight}]'

    def __repr__(self) -> str:
        return f'Transition("{self.name}",tid="{self.tid}",weight={self.weight},' \
               + f'silent={self.silent})'

SILENT_TRANSITION_DEFAULT_NAME='tau'
def silent_transition(name=None,tid=None,weight=None):
    tn = SILENT_TRANSITION_DEFAULT_NAME
    if name:
        tn = name
    tw = 1
    ttid = None
    if tid:
        ttid = tid
    if weight:
        tw = weight
    return Transition(name=tn,weight=tw,tid=ttid,silent=True)

class Arc:
    """
    This is directed arc, which has no name or idenfitier, for Petri net.
    """

    def __init__(self,from_node:Union[Place,Transition],
                 to_node:Union[Place,Transition]):
        self._from_node = from_node
        self._to_node = to_node

    @property
    def from_node(self) -> Union[Place,Transition]:
        return self._from_node

    @property
    def to_node(self) -> Union[Place,Transition]:
        return self._to_node

    def __eq__(self,other) -> bool:
        if type(other) is type(self):
            return self.from_node == other.from_node and \
                    self.to_node  == other.to_node
        return False

    def __hash__(self) -> int:
        return hash((self._from_node,self._to_node))

    def __str__(self) -> str:
        return f'{self.from_node} -> {self.to_node}' 

    def __repr__(self) -> str:
        return f'Arc(from_node={self.from_node.__repr__()},' \
               + f'to_node={self.to_node.__repr__()})'

class LabelledPetriNet:
    """
    This is a data structure for a class of Petri Nets.
    This class consists of places, transitions and directed arcs 
    between them.
    The class contract implies that places and transitions have 
    labels/names and identifiers.Each instance of this class, 
    has a name or title for the net.
    """

    def __init__(self, places:Iterable[Place], transitions:Iterable[Transition],
                 arcs:Iterable[Arc], name:str='Petri net'):
        self._places = set(places)
        self._transitions = set(transitions)
        self._arcs = set(arcs)
        self._name = name

    @property 
    def places(self) -> FrozenSet[Place]:
        return frozenset(self._places)

    @property
    def transitions(self) -> FrozenSet[Transition]:
        return frozenset(self._transitions)

    @property
    def arcs(self) -> FrozenSet[Arc]:
        return frozenset(self._arcs)

    @property
    def name(self) -> str:
        return self._name

    def __eq__(self,other) -> bool:
        if isinstance(other,self.__class__):
            return self._name  == other._name and \
                   self._places == other._places and \
                   self._transitions == other._transitions and \
                   self._arcs   == self._arcs
        return False

    def __repr__(self) -> str:
        repr = "LabelledPetriNet(\n"
        # add places
        repr += "\tplaces=[\n"
        for p in self.places:
            repr += f"\t\t{p.__repr__()},\n"
        repr += "\t],\n"
        # add transitions
        repr += "\ttransitions=[\n"
        for t in self.transitions:
            repr += f"\t\t{t.__repr__()},\n"
        repr += "\t],\n"
        # add arcs
        repr += "\tarcs=[\n"
        for a in self.arcs:
            repr += f"\t\t{a.__repr__()},\n"
        repr += "\t],\n"
        # add name
        repr += f"\tname='{self.name}'\n"
        #close param
        repr += ")"
        return repr

    def __str__(self) -> str:
        _str = "LabelledPetriNet with name of '" + self._name + "'\n"
        _str += "\tPlaces: \n"
        for p in self.places:
            _str += f"\t\t- {p}\n"
        _str += "\tTransitions: \n"
        for t in self.transitions:
            _str += f"\t\t- {t}\n"
        _str += "\tArcs: \n"
        for a in self.arcs:
            _str += f"\t\t- {a}\n"
        return _str
    

class BuildablePetriNet(LabelledPetriNet):
    """
    This class allows for the builder design pattern to be used
    for constructing a petri net. It allows for users to quickly
    add places, transitions and arcs through a single chain of
    method calls. See usage below.

    Usage
    -----
    ```
    # setup elements
    buildable = BuildablePetriNet("dupe_tran_with_id")
    initial_place = Place("I",1)
    ta1 = Transition("a",1)
    ta2 = Transition("a",2)
    tb = Transition("b",3)
    finalPlace = Place("F",2)
    # build net
    buildable.add_place(initialPlace) \\
        .add_transition(ta1) \\
        .add_transition(ta2) \\
        .add_transition(tb) \\
        .add_place(finalPlace) \\
        .add_arc_between(initialPlace, ta1) \\
        .add_arc_between(ta1,finalPlace) \\
        .add_arc_between(initialPlace, ta2) \\
        .add_arc_between(ta2,finalPlace) \\
        .add_arc_between(initialPlace, tb) \\
        .add_arc_between(tb,finalPlace) 
    # get a state of build
    net = buildable.create_net()
    ```
    """
    def __init__(self,label:str=None):
        super(BuildablePetriNet,self).__init__(set(),set(),set(),label)

    def add_place(self,place:Place) -> 'BuildablePetriNet':
        "Adds a place to the net."
        self._places.add(place)
        return self

    def add_transition(self,tran:Transition) -> 'BuildablePetriNet':
        "Adds a tranistion to the net."
        self._transitions.add(tran)
        return self

    def add_arc(self,arc:Arc)-> 'BuildablePetriNet':
        "Adds an arc to the net."
        self._arcs.add(arc)
        return self

    def add_arc_between(self,from_node:Union[Place,Transition],
                             to_node:Union[Place,Transition]
                        ) -> 'BuildablePetriNet' :
        "Constructs an arc between the given nodes and adds it to the net."
        self.add_arc(Arc(from_node,to_node))
        return self

    def create_net(self) -> LabelledPetriNet:
        "Returns a Petri net of the current built state"
        return eval(self.__repr__())

    def __eq__(self,other):
        if isinstance(other,LabelledPetriNet):
            return self._name  == other._name and \
                   self._places == other._places and \
                   self._transitions == other._transitions and \
                   self._arcs   == self._arcs
        return False

class PetriNetDOTFormatter:
    """
    This class creates a dot (graphviz) structure for displaying. Used as internal
    machinery for exporting to a dot file.
    """
    def __init__(self,pn:LabelledPetriNet,font:str='SimSun'):
        self._pn = pn
        self._font = font
        self._nodemap = {}
        self._default_height = 0.2

    def transform_transition(self,tran,ti) -> str:
        fstr = 'n{} [shape="box",margin="0, 0.1",label="{} {}",style="filled"];\n'
        tl = tran.name if tran.name and tran.name != 'tau' else '&tau;'
        fstr = f'n{str(ti)} [shape="box",margin="0, 0.1",'
        fstr += f'label="{tran.weight}", style="filled",'
        height = self._default_height
        fstr += f'height="{height}", width="{height}"'
        fstr += '];\n'
        return fstr

    def transform_place(self,place,pi) -> str:
        fstr = '{} [shape="circle",label="{}"];\n'
        return fstr.format('n' + str(pi),place.name)

    def transform_arc(self,arc) -> str:
        from_node = self._nodemap[arc.from_node]
        to_node = self._nodemap[arc.to_node]
        return f'n{from_node}->n{to_node}\n'

    def transform_net(self) -> str:
        dotstr = ""
        dotstr += 'digraph G{\n'
        dotstr += f'ranksep=".3"; fontsize="14"; remincross=true; margin="0.0,0.0"; fontname="{self._font}";rankdir="LR";charset=utf8;\n'
        dotstr += 'edge [arrowsize="0.5"];\n'
        dotstr += f'node [height="{self._default_height}",width="{self._default_height}",fontname="{self._font}"'
        dotstr += ',fontsize="14"];\n'
        dotstr += 'ratio=0.4;\n'
        ni = 1
        for pl in self._pn.places:
            ni += 1
            self._nodemap[pl] = ni
            dotstr += self.transform_place(pl,ni)
        for tr in self._pn.transitions:
            ni += 1
            self._nodemap[tr] = ni
            dotstr += self.transform_transition(tr,ni)
        for ar in self._pn.arcs:
            dotstr += self.transform_arc(ar)
        dotstr += '}\n'
        return dotstr

def convert_net_to_dot(net:LabelledPetriNet) -> str:
    return PetriNetDOTFormatter(net).transform_net()

PNML_URL='http://www.pnml.org/version-2009/grammar/pnmlcoremodel',
def convert_net_to_xml(net:LabelledPetriNet) -> ET.Element: 
    """
    Converts a given Petri net to an XML structure that conforms with the pnml
    schema.

    See: http://www.pnml.org/version-2009/grammar/pnmlcoremodel.rng
    """
    root = ET.Element('pnml')
    net_node = ET.SubElement(root,'net', 
            attrib={'type':PNML_URL,
                    'id':net.name} )
    page = ET.SubElement(net_node,'page', id="page1")
    for place in net.places:
        placeNode = ET.SubElement(page,'place', attrib={'id':str(place.pid) } )
        if place.name:
            name_node = ET.SubElement(placeNode,'name')
            text_node = ET.SubElement(name_node,'text')
            text_node.text = place.name
    for tran in net.transitions:
        tranNode = ET.SubElement(page,'transition', 
                        attrib={'id':str(tran.tid) } )
        if tran.name:
            name_node = ET.SubElement(tranNode,'name')
            text_node = ET.SubElement(name_node,'text')
            text_node.text = tran.name
        ts_node = ET.SubElement(tranNode,'toolspecific',
                        attrib={ 'tool':'StochasticPetriNet',
                                 'version':'0.2', 
                                 'invisible': str(tran.silent),
                                 'priority': '1',
                                 'weight' : str(tran.weight),
                                 'distributionType': 'IMMEDIATE'} )
    arcid = 1
    for arc in net.arcs:
        arcNode = ET.SubElement(page,'arc',
                    attrib={'source': str(arc.from_node.nodeId), 
                            'target': str(arc.to_node.nodeId),
                            'id': str(arcid) } )
        arcid += 1
    return root


def convert_net_to_xmlstr(net:LabelledPetriNet) -> str: 
    """
    Converts a given Petri net, to an XML structure, and then returns a string
    representation of the indented XML tree.
    """
    xml = convert_net_to_xml(net)  
    ET.indent( xml ) 
    return ET.tostring(xml,encoding=ENCODING)

def export_net_to_pnml(net:LabelledPetriNet,fname:str) -> None: 
    """
    Converts a given Petri net, to an XML structure conforming to the pnml schema 
    , then writes out the XML to a given file location (fname).

    No checking of file location is done for you.
    """
    xml =  convert_net_to_xml(net)  
    ET.indent( xml ) 
    ET.ElementTree(xml).write(fname,xml_declaration=True, encoding="utf-8")




"""
This module is a implementation of the alpha miner (original).

See the following article for more information about the 
alpha miner process discovery technique:\n
W.M.P. van der Aalst, A.J.M.M. Weijters, and L. Maruster. 
Workflow Mining: Discovering Process Models from Event Logs. 
IEEE Transactions on Knowledge and Data Engineering,
16(9):1128–1142, 2004.
"""
from typing import Set,Tuple,Dict,List,Union
from enum import Enum
from copy import deepcopy,copy

from pmkoalas.simple import EventLog
from pmkoalas.directly import DirectlyFollowPair
from pmkoalas.models.petrinet import LabelledPetriNet

class AlphaRelationType(Enum):
    DF = ">"
    CD = "->"
    NF = "#"
    PD = "||"

class AlphaRelation():
    """
    A helper data class to work out what relation is suitable
    for a row, column in the footprint matrix for the alpha
    miner.

    Supported Relations:
    --------------------
    >  : directly follows\n
    -> : causal dependency\n
    \# : never follows\n
    || : parallel dependency\n
    """

    def __init__(self, src:str, target:str, follows:List[Tuple[str,str]]=None) -> None:
        self.target = target
        self.src = src
        self._follows = [] if follows == None else follows
        self.__members = [src, target]
        self.__hash = hash((target,src))

    def add_follows(self, src:str, follow:str) -> None:
        if (src in self.__members and follow in self.__members):
            self._follows.append((src,follow))

    def relation(self) -> AlphaRelationType:
        "returns the alpha relation between src and target"
        if ((self.src,self.target) in self._follows and 
            (self.target,self.src) in self._follows
        ):
            return AlphaRelationType.PD 
        elif ((self.src,self.target) in self._follows and 
              (self.target,self.src) not in self._follows
        ):
            return AlphaRelationType.CD 
        elif ( (self.src,self.target) in self._follows):
            return AlphaRelationType.DF
        return AlphaRelationType.NF

    # data model functions
    def __str__(self) -> str:
        formatter = "({src} {relation} {target})"
        return formatter.format(src=self.src, target=self.target,
            relation=self.relation().value
        )

    def __repr__(self) -> str:
        formatter = 'AlphaRelation(src="{src}",target="{target}",follow={follows})'
        return formatter.format(src=self.src, target=self.target,
            follows=self._follows
        )

    def __hash__(self) -> int:
        return self.__hash

    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, AlphaRelation)):
            return __o.__hash__() == self.__hash__()
        return False

class AlphaPair():
    """
    Data class helper for place generation.
    """

    def __init__(self, left:Set, right:Set) -> None:
        self.left = left 
        self.right = right 
        ord_left = sorted(list([l for l in left]))
        ord_right = sorted(list([r for r in right]))
        self.__hash = hash(
            tuple([l for l in ord_left] + [r for r in ord_right])
        )

    def can_add_left(self, add:str, 
        matrix:Dict[str,Dict[str,AlphaRelation]]) -> bool:
        "Checks if we can add new input transition"
        # check that add is unseen
        if (add in self.left):
            return False
        # check for cd between new left and old rights
        for right in self.right:
            relation = matrix[add][right]
            if (relation.relation() != AlphaRelationType.CD):
                return False
        # check for nf between add and old lefts
        for left in self.left:
            relation = matrix[add][left]
            if (relation.relation() != AlphaRelationType.NF):
                return False
        return True

    def can_add_right(self, add:str, 
        matrix:Dict[str,Dict[str,AlphaRelation]]) -> bool:
        "Checks if we can add new output transition"
        # check that add is unseen
        if (add in self.right):
            return False
        # check for cd between old lefts to new right
        for left in self.left:
            relation = matrix[left][add]
            if (relation.relation() != AlphaRelationType.CD):
                return False
        # check for nf between add and old rights
        for right in self.right:
            relation = matrix[add][right]
            if (relation.relation() != AlphaRelationType.NF):
                return False
        return True

    def expand_left(self, add:str) -> 'AlphaPair':
        "Creates new pair with expanded left, using given add."
        new_left = deepcopy(self.left)
        new_left.add(add)
        new_right = deepcopy(self.right)
        return AlphaPair(new_left, new_right)

    def expand_right(self, add:str) -> 'AlphaPair':
        "Creates new pair with expanded right, using given add."
        new_left = deepcopy(self.left)
        new_right = deepcopy(self.right)
        new_right.add(add)
        return AlphaPair(new_left, new_right)

    def issubset(self, other:'AlphaPair') -> bool:
        "Checks if other implies this pair"
        if (not self.left.issubset(other.left)):
            return False 
        if (not self.right.issubset(other.right)):
            return False 
        return True

    def __hash__(self) -> int:
        return self.__hash

    def __repr__(self) -> str:
        ord_left = sorted([l for l in self.left])
        ord_right = sorted([r for r in self.right])
        left_str = "{" + str(ord_left)[1:-1] + "}"
        right_str = "{" + str(ord_right)[1:-1] + "}"
        return f"({left_str},{right_str})"

    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, AlphaPair)):
            return __o.__hash__() == self.__hash__()
        return False

class AlphaPlace():
    """
    Data class helper for managing places generated from alpha miner.
    """

    def __init__(self, suffix:str, incoming:Set[str], 
        outgoing:Set[str]) -> None:
        self.name = "P_"+suffix
        self.incoming = deepcopy(incoming)
        self.outgoing = deepcopy(outgoing)
        self.__hash = hash(
            tuple(inc for inc in sorted(list(self.incoming)))
            + tuple(out for out in sorted(list(self.outgoing)))
        )
    
    def __hash__(self) -> int:
        return self.__hash

    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, AlphaPlace)):
            return __o.__hash__() == self.__hash__()
        return False 

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return self.__str__()

class AlphaStartPlace(AlphaPlace):
    """
    Special case for starting place in alpha miner.
    """

    ALPHA_START_HASH = "THISISTHEINITIALPLACE"

    def __init__(self) -> None:
        super().__init__("initial", set(), set())

    def __hash__(self) -> int:
        return hash(tuple(self.ALPHA_START_HASH))

class AlphaSinkPlace(AlphaPlace):
    """
    Special case for sink place in alpha miner.
    """

    ALPHA_END_HASH = "THISISTHESINKPLACE"

    def __init__(self) -> None:
        super().__init__("sink", set(), set())

    def __hash__(self) -> int:
        return hash(tuple(self.ALPHA_END_HASH))

class AlphaTransition():
    """
    Data class helper for managing transitions generated from
    alpha miner.
    """

    def __init__(self,name:str) -> None:
        self.name = name
        self.__hash = hash(tuple(name))

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return self.name

    def __hash__(self) -> int:
        return self.__hash

    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, AlphaTransition)):
            return self.__hash__() == __o.__hash__()
        return False
    
class AlphaFlowRelation():
    """
    Data class helper for managing flow relations generated
    from alpha miner.
    """

    def __init__(self,suffix:str, 
        src:Union[AlphaPlace,AlphaTransition],
        tar:Union[AlphaPlace,AlphaTransition]) -> None:
        self.name = "F_"+suffix
        self.src = src 
        self.tar = tar 
        self.__hash = hash(tuple([src,tar]))

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return f"{self.src} -> {self.tar}"

    def __hash__(self) -> int:
        return self.__hash

    def __eq__(self, __o: object) -> bool:
        if (isinstance(__o, AlphaFlowRelation)):
            return self.__hash__() == __o.__hash__()
        return False

class AlphaMinerInstance():
    """
    This is a helper for the alpha miner using the same 
    parameters. The alpha miner consists of several steps,
    which each steps finds a finite set defined by an event 
    log. The alpha miner algorithm returns the components 
    needed to defined a Petri net, i.e. a set of places, a 
    set of transitions, and a set of flow relations.
    \n
    step one: find the set of transitions (process activities)\n
    step two: find the set of start activities\n
    step three: find the set of end activities\n
    step four: determine all possible places for the net\n
    step five: determine the maximal pairs for the net\n
    step six: \n
    step seven: \n
    \n
    The alpha miner has the following parameters:\n
    Parameters
    ----------    
    min_inst:`int=1`(default)\n
    The minimum number of times a relation must be seen to
    be counted as a alpha relation.
    """

    def __init__(self, min_inst:int=1):
        self._min_inst = min_inst
        self._matrix = None

    def mine_footprint_matrix(self, log:EventLog) -> Dict[
        str,Dict[str,AlphaRelation]]:
        """
        Mines the footprint matrix of the alpha miner relations
        between process activities.
        """
        self._matrix = dict()
        TL = self._step_one(log)
        # set up matrix 
        for src in TL:
            self._matrix[src] = dict()
            for target in TL:
                self._matrix[src][target] = AlphaRelation(
                    src=src,target=target
                )
        # find directly flows lang
        flang = log.directly_follow_relations()
        # update relations
        for col in self._matrix.values():
            for relation in col.values():
                src = relation.src
                target = relation.target
                # test flow relation from src to target
                pair = DirectlyFollowPair(src, target, 1)
                if (flang.contains(pair)):
                    fpair = flang.find(pair)
                    if fpair.frequency() >= self._min_inst:
                        # print(f"adding follows ({src},{target}) to {relation}")
                        relation.add_follows(src, target)
                # test flow relation from target to src
                pair = DirectlyFollowPair(target, src, 1)
                if (flang.contains(pair)):
                    fpair = flang.find(pair)
                    if fpair.frequency() >= self._min_inst:
                        # print(f"adding follows ({target},{src}) to {relation}")
                        relation.add_follows(target, src)
        # make new copy
        out = dict()
        for src in TL:
            out[src] = deepcopy(self._matrix[src])
        return deepcopy(self._matrix)

    def mine_model(self, log:EventLog) -> LabelledPetriNet:
        """
        Mines a petri net based on the given event log, using
        the alpha miner discovery technique.

        Returns a Petri net, i.e. a set of places, a set of 
        transitions and a set of flow relations.
        """
        self._matrix = None
        # find the essential elements 
        matrix = self.mine_footprint_matrix(log)
        ### the set of transitions
        TL = self._step_one(log)
        # others 
        TI = self._step_two(log)
        TO = self._step_three(log)
        XL = self._step_four(log)
        YL = self._step_five(log, XL)
        # find the petri net components 
        ### the set of places
        PL = self._step_six(log, YL) 
        ### the set of flow relations
        FL = self._step_seven(log, PL, TI, TO) 
        TL = set([ AlphaTransition(t) for t in TL])
        self._matrix = None
        return LabelledPetriNet(PL, TL, FL)

    def _step_one(self, log:EventLog) -> Set[str]:
        """
        Returns a set of all process activities seen in the 
        event log.
        """
        return log.seen_activities()

    def _step_two(self, log:EventLog) -> Set[str]:
        """
        Returns a set of all start process activities seen in 
        the event log.
        """
        return log.seen_start_activities()

    def _step_three(self, log:EventLog) -> Set[str]:
        """
        Returns a set of all end process activities seen in 
        the event log.
        """
        return log.seen_end_activities()

    def _step_four(self, log:EventLog) -> Set:
        """
        Returns a set of pairs, where each pair describes a 
        set of transitions that have a causal dependency (->) 
        with all the transitions in the second set. Futhermore, 
        each member of each set, is never seen follows (#) 
        another member from their set.
        """
        if (self._matrix == None):
            self._matrix = self.mine_footprint_matrix(log)
        pairs:Set[AlphaPair] = set()
        # build all pair iteratively, start within |A| = 1
        TL = self._step_one(log)
        for src in TL:
            col = self._matrix[src]
            # check that src has never follows with itself
            if (col[src].relation() != AlphaRelationType.NF):
                continue
            for tar in TL:
                relation = col[tar]
                # check for casual dependency
                if (relation.relation() == AlphaRelationType.CD):
                    # add to base
                    pairs.add( 
                        AlphaPair(set([src]),set([tar]))
                    )
        # keep expanding until pairs does not grow
        last_size = 0
        next_size = len(pairs)
        while last_size != next_size:
            last_size = len(pairs)
            # to expand we do the following for each pair in pairs
            # we check for adding a new left for each pair
            # then check for adding a new right for each pair
            new_pairs = deepcopy(pairs)
            for t in TL:
                for pair in pairs:
                    if (pair.can_add_left(t, self._matrix)):
                        new_pairs.add(pair.expand_left(t))
                    if (pair.can_add_right(t, self._matrix)):
                        new_pairs.add(pair.expand_right(t))
            pairs.update(new_pairs)
            # update end condition
            next_size = len(pairs)
        return pairs

    def _step_five(self, log:EventLog, XL:Set) -> Set:
        """
        Returns all maximal pairs from XL, reducing the number
        of places in the Petri net.
        """
        maximal_pairs = set()
        for pair in XL:
            issubset = False
            for opair in XL:
                if pair == opair:
                    continue 
                issubset = issubset or pair.issubset(opair)
                if (issubset):
                    break
            if (not issubset):
                maximal_pairs.add(deepcopy(pair))
        return maximal_pairs 

    def _step_six(self, log:EventLog, YL:Set[AlphaPair]) -> Set:
        """
        Returns a set of alpha places for the Petri net, including 
        an initial place and a sink place.
        """
        places = set()
        for y in YL:
            places.add(AlphaPlace(str(y), y.left, y.right))
        places.add(AlphaStartPlace())
        places.add(AlphaSinkPlace())
        return places

    def _step_seven(self, log:EventLog, PL:Set[AlphaPlace], 
        TI:Set[str], TO:Set[str]) -> Set:
        """
        Returns a set of alpha flow relations, which are directed
        arcs from source to some target.
        """
        flows = set()
        for place in PL:
            if isinstance(place, AlphaSinkPlace):
                for out in TO:
                    flows.add(AlphaFlowRelation(str(len(flows)+1), 
                        AlphaTransition(out),
                        place
                    )) 
            elif isinstance(place, AlphaStartPlace):
                for inc in TI:
                    flows.add(AlphaFlowRelation(str(len(flows)+1), 
                        place,
                        AlphaTransition(inc)
                    )) 
            else:
                for inc in place.incoming:
                    flows.add(AlphaFlowRelation(str(len(flows)+1),
                        AlphaTransition(inc),
                        place 
                    ))
                for out in place.outgoing:
                    flows.add(AlphaFlowRelation(str(len(flows)+1),
                        place, 
                        AlphaTransition(out)
                    ))
        return flows

    




import os.path
import shutil
import subprocess
import tempfile
from logging import info

DFENC='utf-8'

def dot_to_img(dot:str,dotf:str,imgf:str,outformat:str,dfenc:str=DFENC):
    with open(dotf,'w',encoding=dfenc) as outf:
        outf.write(dot)
        outf.write('\n')
    tfd, tfnamed = tempfile.mkstemp()
    tfi, tfnamei = tempfile.mkstemp()
    shutil.copy(dotf,tfnamed)
    cmdout = subprocess.run(['dot','-T' + outformat,'-o' + tfnamei, tfnamed])
    shutil.copy(tfnamei,imgf)

def export_DOT_to_image(vard:str,oname:str,dotStr:str,outformat:str='png'):
    dotf = os.path.join(vard,oname+'.dot')
    imgpnf = os.path.join(vard,oname+'_pn.' + outformat)
    info(f"Generating diagram {oname} ... ")
    dot_to_img(dotStr,dotf,imgpnf,outformat)





'''
Allows the creation of Petri nets with short one line ASCII sketches, for 
example:
initialPlace -> [transition1] -> mp -> [transition2] -> finalPlace

Larger nets can be created with multiple invocations. Existing nodes will be 
looked up by label. 

Weighted transitions without weights, as in {b}, are defaulted to weight 1.0.

Nodes with duplicate labels can be specified using a [tranLabel__id] syntax. 

Example uses::
    net1 = pnfrag.create_net()
    net2 = pnfrag.parse_net_fragments("tester_net",
                "I -> [A] -> p1 -> [tau] -> F",
                "I -> [A] -> p1 -> [B] -> p2 -> [E] -> p3 -> [G] -> F",
                "I -> [A] -> p1 -> [C] -> p2 -> [E] -> p3 -> [H] -> F",
                "I -> [A] -> p1 -> [D] -> p2 -> [E] -> p3 -> [K] -> F")
    parser = pnfrag.PetriNetFragmentParser()
    net3 = parser.create_net()
    parser.add_to_net(net3, "I -> {a__1 3} -> F")
 

Weighted transitions, allowing support of Stochastic Labelled Petri Nets. 

Current limitations: No time distributions or weighted arcs.

 
Grammar

PETRI_ONELINE_NET 	:: PLACE EDGE TRANSITION EDGE PLACE_LED_SUBNET
PLACE_LED_SUBNET  	:: PLACE EDGE TRANSITION EDGE PLACE_LED_SUBNET
PLACE_LED_SUBNET  	:: PLACE 
TRANSITION_SUBNET 	:: TRANSITION EDGE PLACE EDGE TRANSITION_SUBNET
TRANSITION_SUBNET 	:: TRANSITION 
TRANSITION        	:: SIMPLE_TRANSITION || WEIGHTED_TRANSITION
SIMPLE_TRANSITION 	:: TRAN_START TRAN_LABEL TRAN_END
WEIGHTED_TRANSITION     :: WEIGHTED_TRAN_VALUE | WEIGHTED_TRAN_DEFAULT
WEIGHTED_TRAN_VALUE     :: '{' TRAN_LABEL WEIGHT '}'
WEIGHTED_TRAN_DEFAULT   :: '{' TRAN_LABEL '}'
TRAN_LABEL		:: TLABEL || T_ID_LABEL
T_ID_LABEL		:: TLABEL ID_PREFIX ID
TLABEL			:: LABEL || SILENT_LABEL
PLACE             	:: LABEL || P_ID_LABEL
P_ID_LABEL              :: LABEL ID_PREFIX ID
EDGE              	:: '->'
SIMPLE_TRAN_START	:: '['
SIMPLE_TRAN_END		:: ']' 
ID_PREFIX		:: '__'
WEIGHT		 	:: NUM_STR
ID             		:: NUM_STR
NUM_STR			:: numeric string
SILENT_LABEL            :: 'tau'
LABEL             	:: alphanumeric string

This is adapted from PetriNetFragmentParser.java in the prom-helpers library.

See also: 
    - https://github.com/adamburkegh/prom-helpers
    - https://adamburkeware.net/2021/05/20/petri-net-fragments.html

'''

from enum import Enum
import re

from pmkoalas.models.petrinet import *
from logging  import debug


ID_LEXEME = '__'

class TokenInfo(Enum):
    SIMPLE_TRAN_START = "\\[",
    SIMPLE_TRAN_END = "\\]",
    WEIGHTED_TRAN_START = "\\{",
    WEIGHTED_TRAN_END = "\\}",
    ID_PREFIX = ID_LEXEME,
    EDGE = "->",
    SILENT_LABEL = "tau",
    LABEL = "[a-zA-Z][a-zA-Z0-9]*",
    WEIGHT = "[0-9]+\\.[0-9]+",
    ID = "[0-9]+",
    TERMINAL = ""

    def __init__(self,strp):
        self._strp = strp
        self._pattern = re.compile(strp)

    @property
    def pattern(self):
        return self._pattern

TOKEN_LEX_VALUES = [toki for toki in TokenInfo] 
TOKEN_LEX_VALUES.remove(TokenInfo.TERMINAL)


class Token:
    def __init__(self,token_info,token_str):
        self.token_info = token_info
        self.token_str = token_str

class ParseException(Exception):
    pass

def create_net(net_title, net_text) -> LabelledPetriNet:
    return PetriNetFragmentParser().create_net(net_title,net_text)

def parse_net_fragments(net_title:str, *fragments:str) -> LabelledPetriNet:
    """
    Given a set of fragments following the gammar mentioned below, this function
    returns a Petri net that contains all the fragments.

    See the module documentation for a full grammar.
    """
    parser = PetriNetFragmentParser()
    if (len(fragments) > 0):
        net = parser.create_net(net_title, fragments[0])
        for frag in fragments[1:]:
            parser.add_to_net(net, frag)
    return net


class PetriNetFragmentParser:
    def __init__(self):
        self.init()

    def init(self):
        self.tokens = []
        self.ctid = 0
        self.label_lookup = {}
        self.id_lookup = {}

    def create_net(self,net_title:str, net_text:str) -> BuildablePetriNet:
        net = BuildablePetriNet(label=net_title)
        self.init()
        self.add_to_net(net,net_text)
        return net

    def add_to_net(self, net:LabelledPetriNet, net_text: str):
        self.tokenize(net_text)
        self.net = net
        self.parse()

    def tokenize(self,tstr: str):
        self.tokens = []
        ctstr = tstr.strip()
        while (ctstr != ""):
            match = False
            ctstr = ctstr.strip()
            for tokinfo in TOKEN_LEX_VALUES:
                mo = tokinfo.pattern.match(ctstr)
                if mo:
                    match = True
                    tok = mo.group().strip()
                    debug( f'{tokinfo}::{tok}')
                    self.tokens.append( Token(tokinfo,tok) )
                    ctstr = ctstr.replace(mo.group(),"",1)
                    break
            if not match:
                raise ParseException("Unexpected character in input:\""  \
                                    + ctstr + "\"")

    def parse(self):
        self.lookahead = self.tokens[0]
        self.petri_one_line_net()
        if(self.lookahead.token_info != TokenInfo.TERMINAL):
            raise ParseException(f"Unexpected symbol {self.lookahead} found")

    def check_existing_node(self,label: str, node_type):
        if not label in self.label_lookup:
            return None
        n = self.label_lookup[label]
        if node_type != type(n):
            raise ParseException(f"New node {label} duplicates existing node of wrong type")
        return n

    def check_existing_node_by_id(self, nid, node_type):
        if not nid in self.id_lookup:
            return None
        n = self.id_lookup[nid]
        if node_type != type(n):
            raise ParseException(f"New node {nid} duplicates existing node of wrong type")
        return n

    def check_existing_place(self, label: str):
        return self.check_existing_node(label,Place)

    def check_existing_place_by_id(self,pid):
        return self.check_existing_node_by_id(pid,Place)

    def check_existing_transition(self,label: str):
        return self.check_existing_node(label,Transition)

    def check_existing_transition_by_id(self,tran_id):
        return self.check_existing_node_by_id(tran_id,Transition)

    def petri_one_line_net(self):
        p1 = self.place()
        self.edge()
        transition = self.transition()
        self.edge()
        p2 = self.place_led_subnet()
        self.readd_arc(p1, transition)
        self.readd_arc(transition, p2)

    def place(self):
        label = self.lookahead.token_str
        self.next_token()
        pid = None
        place = None
        if self.lookahead.token_info == TokenInfo.ID_PREFIX:
            self.next_token()
            pid = self.id()
            self.next_token()
        if pid:
            place = self.check_existing_place_by_id(pid)
        else:
            place = self.check_existing_place(label)
        if (place is None):
            if not pid:
                pid = self.next_id()
            place = Place(label,pid=pid)
            self.net.add_place(place)
            self.label_lookup[label] = place
        return place;

    def edge(self):
        if self.lookahead.token_info != TokenInfo.EDGE:
            self.token_error(TokenInfo.EDGE)
        self.next_token()

    def transition(self):
        transition = None
        if self.lookahead.token_info == TokenInfo.SIMPLE_TRAN_START: 
            transition = self.simple_transition()
        if self.lookahead.token_info == TokenInfo.WEIGHTED_TRAN_START: 
            transition = self.weighted_value_transition()
        self.next_token()
        return transition

    def simple_transition(self):
        transition = None
        self.next_token()
        label = ''
        tran_id = None
        silent_tran = False
        if self.lookahead.token_info == TokenInfo.LABEL:
            label = self.tran_label()
            self.next_token()
        else: 
            if self.lookahead.token_info == TokenInfo.SILENT_LABEL:
                label = self.tran_label()
                silent_tran = True
                self.next_token()
            else:
                raise ParseException(
                        f"Expected label, but found {self.lookahead.token_str}")
        if self.lookahead.token_info == TokenInfo.SIMPLE_TRAN_END:
            transition = self.check_existing_transition(label)
        else:
            if self.lookahead.token_info == TokenInfo.ID_PREFIX:
                transition = None
                self.next_token()
                tran_id = self.id()
                self.next_token()
                if self.lookahead.token_info != TokenInfo.SIMPLE_TRAN_END:
                    self.token_error(TokenInfo.SIMPLE_TRAN_END,
                                    TokenInfo.ID_PREFIX)
            else:
                self.token_error(TokenInfo.SIMPLE_TRAN_END,
                                TokenInfo.ID_PREFIX)
        if not tran_id is None:
            transition = self.check_existing_transition_by_id(tran_id)
        if not transition:
            if tran_id:
                transition = Transition(label,tid=tran_id,silent=silent_tran)
                self.id_lookup[tran_id] = transition
            else:
                transition = Transition(label,tid=self.next_id(),
                                        silent=silent_tran)
                self.label_lookup[label] = transition
            self.net.add_transition(transition)
        return transition

    def weighted_value_transition(self):
        transition = None
        self.next_token()
        label = ''
        tran_id = None
        silent_tran = False
        weight = 1.0
        if self.lookahead.token_info == TokenInfo.LABEL:
            label = self.tran_label()
            self.next_token()
        else:
            if self.lookahead.token_info == TokenInfo.SILENT_LABEL:
                label = self.tran_label()
                silent_tran = True
                self.next_token()
            else:
                raise ParseException(
                        f"Expected label, but found {self.lookahead.token_str}")

        if self.lookahead.token_info == TokenInfo.ID_PREFIX:
            transition = None
            self.next_token()
            tran_id = self.id()
            self.next_token()
        if self.lookahead.token_info == TokenInfo.WEIGHT:
            weight = self.weight()
            self.next_token()
        if self.lookahead.token_info == TokenInfo.WEIGHTED_TRAN_END:
            transition = self.check_existing_transition(label)
        else:
            self.token_error(TokenInfo.WEIGHTED_TRAN_END,
                            TokenInfo.ID_PREFIX)
        if not transition:
            if tran_id:
                transition = Transition(label,tid=tran_id,silent=silent_tran,
                                    weight=weight)
                self.id_lookup[tran_id] = transition
            else:
                transition = Transition(label,tid=self.next_id(),
                                         silent=silent_tran, weight=weight)
                self.label_lookup[label] = transition
            self.net.add_transition(transition)
        return transition


    def tran_label(self):
        return self.lookahead.token_str

    def place_led_subnet(self):
        head = self.place()
        if self.lookahead.token_info == TokenInfo.EDGE:
            self.edge()
            tran = self.transition()
            self.edge()
            tail = self.place_led_subnet()
            self.readd_arc(head,tran)
            self.readd_arc(tran,tail)
        return head

    def id(self):
        return int(self.lookahead.token_str)

    def weight(self):
        return float(self.lookahead.token_str)

    def token_error(self,*tokens):
        raise ParseException(
                f"Expected one of {tokens} but found {self.lookahead.token_str}")


    def next_token(self):
        self.tokens.pop(0)
        # at the end of input we return an epsilon token
        if not self.tokens:
            self.lookahead = Token(TokenInfo.TERMINAL, "")
        else:
            self.lookahead = self.tokens[0]

    def readd_arc(self,from_node,to_node):
        self.net.add_arc_between(from_node,to_node)

    def next_id(self):
        self.ctid += 1
        return self.ctid

